"""
The Condor plugin.  Allows submission to
`HTCondor <http://research.cs.wisc.edu/htcondor/>`_.

Note: Condor was renamed to HTCondor in 2012.
"""
import asyncio
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime, timedelta, UTC
import enum
import importlib
import logging
import os
from pathlib import Path
import shutil
import stat
import subprocess
import time
from typing import NamedTuple

import htcondor  # type: ignore

from iceprod.core.config import Task
from iceprod.core.exe import WriteToScript, Transfer
from iceprod.server.config import IceProdConfig
from iceprod.server import grid
from iceprod.server.util import str2datetime

logger = logging.getLogger('condor')


def check_call_clean_env(*args, **kwargs):
    env = os.environ.copy()
    del env['LD_LIBRARY_PATH']
    kwargs['env'] = env
    return subprocess.check_call(*args, **kwargs)


def check_output_clean_env(*args, **kwargs):
    env = os.environ.copy()
    del env['LD_LIBRARY_PATH']
    kwargs['env'] = env
    return subprocess.check_output(*args, **kwargs)


@enum.unique
class JobStatus(enum.Enum):
    IDLE = enum.auto()       # job is waiting in the queue
    RUNNING = enum.auto()    # job is running
    FAILED = enum.auto()     # job needs cleanup
    COMPLETED = enum.auto()  # job is out of the queue

    @staticmethod
    def from_condor_status(num):
        match num:
            case 0 | 1: return JobStatus.IDLE
            case 2 | 6: return JobStatus.RUNNING
            case 3 | 4: return JobStatus.COMPLETED
            case _: return JobStatus.FAILED


JOB_EVENT_STATUS_TRANSITIONS = {
    htcondor.JobEventType.SUBMIT: JobStatus.IDLE,
    htcondor.JobEventType.JOB_STAGE_IN: JobStatus.RUNNING,
    htcondor.JobEventType.JOB_STAGE_OUT: JobStatus.RUNNING,
    htcondor.JobEventType.FILE_TRANSFER: JobStatus.RUNNING,
    htcondor.JobEventType.EXECUTE: JobStatus.RUNNING,
    htcondor.JobEventType.JOB_EVICTED: JobStatus.IDLE,
    htcondor.JobEventType.JOB_UNSUSPENDED: JobStatus.IDLE,
    htcondor.JobEventType.JOB_RELEASED: JobStatus.IDLE,
    htcondor.JobEventType.SHADOW_EXCEPTION: JobStatus.IDLE,
    htcondor.JobEventType.JOB_RECONNECT_FAILED: JobStatus.IDLE,
    htcondor.JobEventType.JOB_TERMINATED: JobStatus.FAILED,
    htcondor.JobEventType.JOB_HELD: JobStatus.FAILED,
    htcondor.JobEventType.JOB_SUSPENDED: JobStatus.FAILED,
    htcondor.JobEventType.JOB_ABORTED: JobStatus.FAILED,
}


RESET_CONDOR_REASONS = [
    # condor file transfer plugin failed
    '_condor_stdout: (errno 2) No such file',
    'transfer input files failure',
    'transfer output files failure',
    # resource limits
    'cpu consumption limit exceeded',
    'memory limit exceeded',
    'cgroup memory limit',
    'local storage limit on worker node exceeded',
    'execution time limit exceeded',
    # general retries
    'exceeded max iceprod queue time',
    'job has failed',
    'python-initiated action (by user ice3simusr)',
]


RESET_STDERR_REASONS = [
    # glidein died
    'sigterm',
    'killed',
    # hopefully transient errors
    'bus error (core dumped)',
    'segmentation fault (core dumped)',
    'operation timed out',
    'connection timed out',
    # GPU errors
    'opencl error: could not set up context',
    'opencl error: could build the opencl program',
    # CVMFS errors
    'python: command not found',
    'cannot read file data: Stale file handle',
    'setenv: command not found',
]


def parse_usage(usage: str) -> int:
    """
    Parse HTCondor usage expression

    Example input: "Usr 0 00:00:00, Sys 0 00:00:00"

    Args:
        usage: usage expression

    Returns:
        usage sum in seconds
    """
    total = 0
    for part in usage.split(','):
        _, days, leftovers = part.strip().split(' ')
        hours, minutes, seconds = leftovers.split(':')
        total += int(days)*86400 + int(hours)*3600 + int(minutes)*60 + int(seconds)
    return total


@dataclass(kw_only=True, slots=True)
class CondorJob(grid.GridTask):
    """Holds the job states for an HTCondor cluster."""
    dataset_id: str | None = None
    task_id: str | None = None
    instance_id: str | None = None
    submit_dir: Path | None = None
    status: JobStatus = JobStatus.IDLE
    extra: dict | None = None


class CondorJobId(NamedTuple):
    """Represents an HTCondor job id"""
    cluster_id: int
    proc_id: int

    def __str__(self):
        return f'{self.cluster_id}.{self.proc_id}'


class CondorSubmit:
    """Factory for submitting HTCondor jobs"""
    AD_DEFAULTS = {
        'request_cpus': 1,
        'request_gpus': 'UNDEFINED',
        'request_memory': 1000,
        'request_disk': 1000000,
        '+OriginalTime': 3600,
        'requirements': '',
        'PreCmd': 'UNDEFINED',
        'PreArguments': 'UNDEFINED',
        'transfer_input_files': [],
        'transfer_output_files': [],
        'transfer_output_remaps': [],
    }

    _GENERIC_ADS = ['Iwd', 'IceProdDatasetId', 'IceProdTaskId', 'IceProdTaskInstanceId', 'MATCH_EXP_JOBGLIDEIN_ResourceName']
    AD_INFO = [
        'RemotePool', 'RemoteHost',
        'HoldReason', 'RemoveReason', 'Reason', 'MachineAttrGLIDEIN_Site0',
    ] + _GENERIC_ADS
    AD_PROJECTION_QUEUE = ['JobStatus', 'RemotePool', 'RemoteHost'] + _GENERIC_ADS
    AD_PROJECTION_HISTORY = [
        'JobStatus', 'ExitCode', 'RemoveReason', 'LastHoldReason', 'CpusUsage', 'RemoteSysCpu', 'RemoteUserCpu',
        'GpusUsage', 'ResidentSetSize_RAW', 'DiskUsage_RAW', 'LastRemoteWallClockTime',
        'LastRemoteHost', 'LastRemotePool', 'MachineAttrGLIDEIN_Site0',
    ] + _GENERIC_ADS

    def __init__(self, cfg: IceProdConfig, submit_dir: Path, credentials_dir: Path):
        self.cfg = cfg
        self.submit_dir = submit_dir
        self.credentials_dir = credentials_dir
        self.condor_schedd = htcondor.Schedd()

        submit_dir.mkdir(parents=True, exist_ok=True)
        self.precmd = submit_dir / 'pre.py'
        self.precmd.write_text((importlib.resources.files('iceprod.server')/'data'/'condor_input_precmd.py').read_text())
        self.precmd.chmod(0o777)
        self.transfer_plugins = self.condor_plugin_discovery()
        logger.info('transfer plugins installed: %s', list(self.transfer_plugins.keys()))

        self.default_container = self.cfg['queue'].get('default_container', 'Undefined')

    def condor_plugin_discovery(self):
        """Find all available HTCondor transfer plugins, and copy them to the submit_dir"""
        src_dir = importlib.resources.files('iceprod.server')/'data'/'condor_transfer_plugins'
        if src_dir.is_dir():
            dest_dir = self.submit_dir / 'transfer_plugins'
            shutil.copytree(src_dir, dest_dir, dirs_exist_ok=True)
            ret = {}
            for p in dest_dir.iterdir():
                p.chmod(0o777)
                logger.debug('transfer plugin %s', p)
                for line in subprocess.run([str(p), '-classad'], capture_output=True, text=True, check=True).stdout.split('\n'):
                    logger.debug('transfer plugin output: %s', line)
                    if line.startswith('SupportedMethods'):
                        ret[line.split('=')[-1].strip(' "')] = str(p)
        return ret

    @staticmethod
    def condor_os_container(os_arch):
        """Convert from OS_ARCH to container image"""
        if isinstance(os_arch, list):
            os_arch = os_arch[0]
        os_arch = os_arch.rsplit('_',2)[0].rsplit('.',1)[0]
        if 'RHEL' not in os_arch:
            raise Exception('unknown OS_ARCH specified')
        version = os_arch.split('_')[-1]
        container = f'/cvmfs/singularity.opensciencegrid.org/opensciencegrid/osgvo-el{version}:latest'
        return container

    @staticmethod
    def condor_resource_reqs(task: Task):
        """Convert from Task requirements to HTCondor requirements"""
        ads = {}
        requirements = []
        if 'cpu' in task.requirements and task.requirements['cpu']:
            ads['request_cpus'] = task.requirements['cpu']
        if 'gpu' in task.requirements and task.requirements['gpu']:
            ads['request_gpus'] = task.requirements['gpu']
            requirements.append('GPUs_Capability >= 6.1')
        if 'memory' in task.requirements and task.requirements['memory']:
            ads['request_memory'] = int(task.requirements['memory']*1000)
        if 'disk' in task.requirements and task.requirements['disk']:
            # add 1G spare for log files and other misc
            ads['request_disk'] = int(task.requirements['disk']*1000000+1000000)
        if 'time' in task.requirements and task.requirements['time']:
            ads['+OriginalTime'] = int(task.requirements['time']*3600)
            requirements.append('TargetTime > OriginalTime')
        if requirements:
            ads['requirements'] = '('+')&&('.join(requirements)+')'
        return ads

    def condor_infiles(self, infiles):
        """Convert from set[Data] to HTCondor classads for input files"""
        files = []
        mapping = []
        x509_proxy = self.cfg['queue'].get('x509proxy', None)
        if x509_proxy:
            files.append(x509_proxy)
        for infile in infiles:
            if infile.url.startswith('gsiftp:') and not x509_proxy:
                raise RuntimeError('need x509 proxy for gridftp!')
            if infile.url[0] == '/':
                url = infile.url
            else:
                if infile.transfer == Transfer.MAYBE:
                    url = 'iceprod-plugin://maybe-' + infile.url
                else:
                    url = 'iceprod-plugin://true-' + infile.url
                basename = Path(infile.url).name
                if basename != infile.local:
                    url += '?mapping='+infile.local
                    # mapping.append((basename,infile.local))
            files.append(url)
        ads = {}
        if mapping:
            ads['PreCmd'] = f'"{self.precmd.name}"'
            ads['PreArguments'] = '"'+' '.join(f"'{k}'='{v}'" for k,v in mapping)+'"'
            files.append(str(self.precmd))
        if files:
            ads['transfer_input_files'] = files
        return ads

    def condor_outfiles(self, outfiles):
        """Convert from set[Data] to HTCondor classads for output files"""
        files = []
        mapping = []
        for outfile in outfiles:
            files.append(outfile.local)
            if outfile.transfer == Transfer.MAYBE:
                url = 'iceprod-plugin://maybe-' + outfile.url
            else:
                url = outfile.url
            mapping.append((outfile.local, url))
        ads = {}
        ads['transfer_output_files'] = files
        ads['transfer_output_remaps'] = ';'.join(f'{name} = {url}' for name,url in mapping)
        return ads

    def create_submit_dir(self, task: Task, jel_dir: Path) -> Path:
        """
        Create the submit dir
        """
        path = jel_dir / task.task_id
        i = 1
        while path.exists():
            path = jel_dir / f'{task.task_id}_{i}'
            i += 1
        path.mkdir(parents=True)
        return path

    async def submit(self, tasks: list[Task], jel: Path) -> dict[CondorJobId, CondorJob]:
        """
        Submit multiple jobs to Condor as a single batch.

        Assumes that the resource requirements are identical.

        Args:
            tasks: IceProd Tasks to submit
            jel: common job event log

        Returns:
            dict of new jobs
        """
        jel_dir = jel.parent
        transfer_plugin_str = ';'.join(f'{k}={v}' for k,v in self.transfer_plugins.items())
        submitfile = f"""
output = $(initialdir)/condor.out
error = $(initialdir)/condor.err
log = {jel}

notification = never
job_ad_information_attrs = {" ".join(self.AD_INFO)}
batch_name = Dataset {tasks[0].dataset.dataset_num}

+IsIceProdJob = True
+IceProdSite = "{self.cfg["queue"].get("site", "unknown")}"
+IceProdDatasetId = $(datasetid)
+IceProdDataset = $(dataset)
+IceProdJobId = $(jobid)
+IceProdJobIndex = $(jobindex)
+IceProdTaskId = $(taskid)
+IceProdTaskName = $(taskname)
+IceProdTaskInstanceId = $(taskinstance)

request_cpus = $(cpus)
request_gpus = $(gpus)
request_memory = $(memory)
request_disk = $(disk)
+OriginalTime = $(time)
+TargetTime = (!isUndefined(Target.PYGLIDEIN_TIME_TO_LIVE) ? Target.PYGLIDEIN_TIME_TO_LIVE : Target.TimeToLive)
requirements = $($(reqs))
+SingularityImage= $(container)

transfer_plugins = {transfer_plugin_str}
when_to_transfer_output = ON_EXIT
should_transfer_files = YES
infiles_expr = replaceall(";", $(infiles), ",")
transfer_input_files = $STRING(infiles_expr)
+PreCmd = $(prec)
+PreArguments = $(prea)
outfiles_expr = replaceall(";", $(outfiles), ",")
transfer_output_files = $STRING(outfiles_expr)
transfer_output_remaps = $(outremaps)

"""

        for k,v in tasks[0].get_task_config()['batchsys'].get('condor', {}).items():
            if k.lower() != 'requirements':
                submitfile += f'+{k} = {v}\n'

        reqs = ''
        for k,v in self.cfg['queue'].get('batchopts', {}).items():
            if k.lower() == 'requirements':
                reqs = v
            else:
                submitfile += f'+{k} = {v}\n'

        jobset = []
        for task in tasks:
            submit_dir = self.create_submit_dir(task, jel_dir)
            s = WriteToScript(task=task, workdir=submit_dir)
            executable = await s.convert(transfer=True)
            logger.debug('running task with exe %r', executable)

            ads = self.AD_DEFAULTS.copy()
            ads.update(self.condor_infiles(s.infiles))
            ads.update(self.condor_outfiles(s.outfiles))
            ads.update(self.condor_resource_reqs(task))

            container = task.get_task_config().get('container')
            if not container:
                if os_arch := task.requirements.get('os'):
                    container = self.condor_os_container(os_arch)
                else:
                    container = self.default_container
            if container != 'Undefined':
                container = f'"{container}"'

            reqs2 = reqs
            for k,v in tasks[0].get_task_config()['batchsys'].get('condor', {}).items():
                if k.lower() == 'requirements':
                    reqs2 = f'({reqs}) && ({v})' if reqs else v
                    break

            if reqs2:
                ads["requirements"] = f'({ads["requirements"]}) && ({reqs2})'
            submitfile += f'reqs{task.task_id} = {ads["requirements"]}\n'
            # stringify everything, quoting the real strings
            jobset.append({
                'datasetid': f'"{task.dataset.dataset_id}"',
                'dataset': f'{task.dataset.dataset_num}',
                'jobid': f'"{task.job.job_id}"',
                'jobindex': f'{task.job.job_index}',
                'taskid': f'"{task.task_id}"',
                'taskname': f'"{task.name}"',
                'taskinstance': f'"{task.instance_id if task.instance_id else ""}"',
                'initialdir': f'{submit_dir}',
                'executable': f'{executable}',
                'cpus': f'{ads["request_cpus"]}',
                'gpus': f'{ads["request_gpus"]}',
                'memory': f'{ads["request_memory"]}',
                'disk': f'{ads["request_disk"]}',
                'time': f'{ads["+OriginalTime"]}',
                'reqs': f'reqs{task.task_id}',
                'container': f'{container}',
                'prec': f'{ads["PreCmd"]}',
                'prea': f'{ads["PreArguments"]}',
                'infiles': f'"{";".join(ads["transfer_input_files"])}"',
                'outfiles': f'"{";".join(ads["transfer_output_files"])}"',
                'outremaps': f'"{ads["transfer_output_remaps"]}"',
            })

        submitfile += '\n\nqueue '+','.join(jobset[0].keys())+' from (\n'
        for job in jobset:
            submitfile += '  '+','.join(job.values())+'\n'
        submitfile += ')\n'

        logger.debug("submitfile:\n%s", submitfile)

        s = htcondor.Submit(submitfile)
        submit_result = self.condor_schedd.submit(s, count=1, itemdata=s.itemdata())

        cluster_id = int(submit_result.cluster())
        ret = {}
        for i,job in enumerate(jobset):
            ret[CondorJobId(cluster_id=cluster_id, proc_id=i)] = CondorJob(
                dataset_id=job['datasetid'].strip('"'),
                task_id=job['taskid'].strip('"'),
                instance_id=job['taskinstance'].strip('"'),
                submit_dir=Path(job['initialdir']),
            )
        return ret

    def get_jobs(self) -> {CondorJobId: CondorJob}:
        """
        Get all jobs currently on the condor queue.
        """
        ret = {}
        for ad in self.condor_schedd.query(
            constraint=f'IceProdSite =?= "{self.cfg["queue"].get("site", "unknown")}"',
            projection=['ClusterId', 'ProcId'] + self.AD_PROJECTION_QUEUE,
        ):
            job_id = CondorJobId(cluster_id=ad['ClusterId'], proc_id=ad['ProcId'])
            submit_dir = None
            if s := ad.get('Iwd'):
                submit_dir = Path(s)
            status = JobStatus.IDLE
            if s := ad.get('JobStatus'):
                status = JobStatus.from_condor_status(s)
            job = CondorJob(
                dataset_id=ad.get('IceProdDatasetId'),
                task_id=ad.get('IceProdTaskId'),
                instance_id=ad.get('IceProdTaskInstanceId'),
                submit_dir=submit_dir,
                status=status,
                extra=ad,
            )
            ret[job_id] = job
        return ret

    def get_history(self, since=None) -> {CondorJobId: CondorJob}:
        """
        Get all jobs currently on the condor history.
        """
        ret = {}
        for ad in self.condor_schedd.history(
            constraint=f'IceProdSite =?= "{self.cfg["queue"].get("site", "unknown")}"',
            projection=['ClusterId', 'ProcId'] + self.AD_PROJECTION_HISTORY,
            since=f'CompletionDate<{since}' if since else None,
        ):
            job_id = CondorJobId(cluster_id=ad['ClusterId'], proc_id=ad['ProcId'])
            if s := ad.get('Iwd'):
                submit_dir = Path(s)
            status = JobStatus.IDLE
            if s := ad.get('JobStatus'):
                status = JobStatus.from_condor_status(s)
            job = CondorJob(
                dataset_id=ad.get('IceProdDatasetId'),
                task_id=ad.get('IceProdTaskId'),
                instance_id=ad.get('IceProdTaskInstanceId'),
                submit_dir=submit_dir,
                status=status,
                extra=ad,
            )
            ret[job_id] = job
        return ret

    def remove(self, job_id: str | CondorJobId, reason: str | None = None):
        """
        Remove a job from condor.

        Args:
            job_id: condor job id
            reason: reason for removal
        """
        logger.info('removing job %s', job_id)
        self.condor_schedd.act(htcondor.JobAction.Remove, str(job_id), reason=reason)


class Grid(grid.BaseGrid):
    """HTCondor grid plugin"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.jobs = {}
        self.jels = {str(filename): htcondor.JobEventLog(str(filename)).events(0) for filename in self.submit_dir.glob('*/*.jel')}
        self.submitter = CondorSubmit(self.cfg, submit_dir=self.submit_dir, credentials_dir=self.credentials_dir)

        # save last event.timestamp, on restart only process >= timestamp
        self.last_event_timestamp = 0.
        self.load_timestamp()

    def load_timestamp(self):
        timestamp_path = self.submit_dir / 'last_event_timestamp'
        if timestamp_path.exists():
            with timestamp_path.open('r') as f:
                self.last_event_timestamp = float(f.read().strip())

    def save_timestamp(self):
        timestamp_path = self.submit_dir / 'last_event_timestamp'
        with timestamp_path.open('w') as f:
            f.write(str(self.last_event_timestamp))

    async def run(self, forever=True):
        # initial job load
        try:
            await self.check()
            self.save_timestamp()
        except Exception:
            logger.warning('failed to check', exc_info=True)
        check_time = time.monotonic()
        try:
            await self.wait(timeout=0)
        except Exception:
            logger.warning('failed to wait', exc_info=True)

        logger.info('active JELs: %r', list(self.jels.keys()))

        while True:
            start = time.monotonic()
            try:
                await self.submit()
            except Exception:
                logger.warning('failed to submit', exc_info=True)
            wait_time = max(0, self.cfg['queue']['submit_interval'] - (time.monotonic() - start))
            try:
                await self.wait(timeout=wait_time)
            except Exception:
                logger.warning('failed to wait', exc_info=True)

            now = time.monotonic()
            if now - check_time >= self.cfg['queue']['check_time']:
                try:
                    await self.check()
                    self.save_timestamp()
                except Exception:
                    logger.warning('failed to check', exc_info=True)
                check_time = now

            if not forever:
                break

    # Submit to Condor #

    async def submit(self):
        num_to_submit = self.get_queue_num()
        logger.info("Attempting to submit %d tasks", num_to_submit)
        tasks = await self.get_tasks_to_queue(num_to_submit)
        cur_jel = self.get_current_JEL()

        # split into datasets and task types
        tasks_by_dataset = defaultdict(list)
        for task in tasks:
            tasks_by_dataset[f'{task.dataset.dataset_num}-{task.name}'].append(task)
        for key in tasks_by_dataset:
            tasks = tasks_by_dataset[key]
            try:
                ret = await self.submitter.submit(tasks, cur_jel)
                self.jobs.update(ret)
            except Exception as e:
                logger.warning('submit failed for dataset %s', key, exc_info=True)
                async with asyncio.TaskGroup() as tg:
                    for task in tasks:
                        j = CondorJob(dataset_id=task.dataset.dataset_id, task_id=task.task_id, instance_id=task.instance_id)
                        tg.create_task(self.task_reset(j, reason=f'HTCondor submit failed: {e}'))

    def get_queue_num(self) -> int:
        """Determine how many tasks to queue."""
        counts = {s: 0 for s in JobStatus}
        for job in self.jobs.values():
            counts[job.status] += 1

        idle_jobs = counts[JobStatus.IDLE]
        logger.info('idle jobs: %r', idle_jobs)
        processing_jobs = counts[JobStatus.RUNNING]
        logger.info('processing jobs: %r', processing_jobs)
        queue_tot_max = self.cfg['queue']['max_total_tasks_on_queue'] - idle_jobs - processing_jobs
        queue_idle_max = self.cfg['queue']['max_idle_tasks_on_queue'] - idle_jobs
        queue_interval_max = self.cfg['queue']['max_tasks_per_submit']
        queue_num = max(0, min(queue_tot_max, queue_idle_max, queue_interval_max))
        return queue_num

    def get_current_JEL(self) -> Path:
        """
        Get the current Job Event Log, possibly creating a new one
        for every hour.

        Returns:
            Path: filename to current JEL
        """
        day = datetime.now(UTC).strftime('%Y-%m-%dT%H')
        day_submit_dir = self.submit_dir / day
        if not day_submit_dir.exists():
            day_submit_dir.mkdir(mode=0o700, parents=True)
        cur_jel = day_submit_dir / 'jobs.jel'
        if not cur_jel.exists():
            cur_jel.touch(mode=0o600)
        cur_jel_str = str(cur_jel)
        if cur_jel_str not in self.jels:
            self.jels[cur_jel_str] = htcondor.JobEventLog(cur_jel_str).events(0)
        return cur_jel

    # JEL processing #

    async def wait(self, timeout):
        """
        Wait for jobs to complete from the Job Event Logs.

        Args:
            timeout: wait up to N seconds
        """
        start = time.monotonic()

        # make sure we have the latest JEL loaded
        self.get_current_JEL()

        while True:
            for filename, events in self.jels.items():
                try:
                    for event in events:
                        if float(event.timestamp) < self.last_event_timestamp:
                            continue

                        job_id = CondorJobId(cluster_id=event.cluster, proc_id=event.proc)

                        if event.type == htcondor.JobEventType.SUBMIT:
                            self.jobs[job_id] = CondorJob()
                            continue
                        elif job_id not in self.jobs:
                            logger.debug('reloaded job %s', job_id)
                            self.jobs[job_id] = CondorJob()

                        job = self.jobs[job_id]

                        if event.type == htcondor.JobEventType.JOB_AD_INFORMATION:
                            if not job.dataset_id:
                                job.dataset_id = event['IceProdDatasetId']
                                job.task_id = event['IceProdTaskId']
                                job.instance_id = event['IceProdTaskInstanceId']
                                job.submit_dir = Path(event['Iwd'])

                            type_ = event['TriggerEventTypeNumber']
                            if type_ == htcondor.JobEventType.JOB_TERMINATED:
                                logger.info("job %s %s.%s exited on its own", job_id, job.dataset_id, job.task_id)

                                # there's a bug where not all the classads are updated before the event fires
                                # so ignore this and let the cross-check take care of it
                                continue

                                # get stats
                                cpu = event.get('CpusUsage', None)
                                gpu = event.get('GpusUsage', None)
                                memory = event.get('ResidentSetSize_RAW', None)  # KB
                                if memory is None:
                                    memory = event.get('MemoryUsage', None)*1000  # MB
                                disk = event.get('DiskUsage_RAW', None)  # KB
                                if disk is None:
                                    disk = event.get('DiskUsage', None)  # KB
                                time_ = event.get('RemoteWallClockTime', None)  # seconds
                                if time_ is None:
                                    time_ = parse_usage(event.get('RunRemoteUsage', '')) / event.get('RequestCpus', 1)
                                elif cpu is None and time_:
                                    cpu = parse_usage(event.get('RunRemoteUsage', '')) / time_
                                # data_in = event['ReceivedBytes']  # KB
                                # data_out = event['SentBytes']  # KB

                                resources = {}
                                if cpu is not None:
                                    resources['cpu'] = cpu
                                if gpu is not None:
                                    resources['gpu'] = gpu
                                if memory is not None:
                                    resources['memory'] = memory/1000000.
                                if disk is not None:
                                    resources['disk'] = disk/1000000.
                                if time_ is not None:
                                    resources['time'] = time_/3600.

                                success = event.get('ReturnValue', 1) == 0
                                job.status = JobStatus.COMPLETED if success else JobStatus.FAILED

                                stats = {}
                                if site := event.get('MachineAttrGLIDEIN_Site0'):
                                    stats['site'] = site
                                elif site := event.get('MATCH_EXP_JOBGLIDEIN_ResourceName'):
                                    stats['site'] = site

                                reason = None
                                if r := event.get('HoldReason'):
                                    reason = r
                                elif r := event.get('RemoveReason'):
                                    reason = r
                                elif r := event.get('Reason'):
                                    reason = r

                                # finish job
                                await self.finish(job_id, success=success, resources=resources, stats=stats, reason=reason)

                            elif type_ == htcondor.JobEventType.JOB_ABORTED:
                                job.status = JobStatus.FAILED
                                reason = event.get('Reason', None)
                                logger.info("job %s %s.%s removed: %r", job_id, job.dataset_id, job.task_id, reason)

                                # there's a bug where not all the classads are updated before the event fires
                                # so ignore this and let the cross-check take care of it
                                continue

                                await self.finish(job_id, success=False, reason=reason)

                            else:
                                # update status
                                new_status = JOB_EVENT_STATUS_TRANSITIONS.get(type_, None)
                                if new_status is not None and job.status != new_status:
                                    job.status = new_status
                                    if new_status == JobStatus.FAILED:
                                        self.submitter.remove(job_id, reason=event.get('HoldReason', 'Job has failed'))
                                    else:
                                        await self.job_update(job)
                except Exception:
                    logger.warning('error processing condor log', exc_info=True)

            if time.monotonic() - start >= timeout:
                break
            await asyncio.sleep(1)

    async def job_update(self, job: CondorJob):
        """
        Send updated info from the batch system to the IceProd API.

        Must handle dup calls.
        """
        if job.status not in (JobStatus.IDLE, JobStatus.RUNNING):
            logging.warning("unknown job status: %r", job.status)
            return

        try:
            if job.status == JobStatus.IDLE:
                logging.info("task %s.%s is now queued", job.dataset_id, job.task_id)
                await self.task_idle(job)
            elif job.status == JobStatus.RUNNING:
                logging.info("task %s.%s is now processing", job.dataset_id, job.task_id)
                await self.task_processing(job)
        except Exception:
            logging.warning("failed to update task status", exc_info=True)

    async def finish(self, job_id: CondorJobId, success: bool = True, resources: dict | None = None, reason: str | None = None, stats: dict | None = None):
        """
        Run cleanup actions after a batch job completes.

        Must handle dup calls.
        """
        if job_id not in self.jobs:
            logger.debug('dup call: %s not in job dict', job_id)
            return

        if not stats:
            stats = {}
        if resources:
            stats['resources'] = resources

        job = self.jobs[job_id]
        logger.info('finish for condor=%s iceprod=%s.%s', job_id, job.dataset_id, job.task_id)

        stdout = None
        stderr = None
        if job.submit_dir and job.submit_dir.is_dir():
            stdout = job.submit_dir / 'condor.out'
            stderr = job.submit_dir / 'condor.err'

        # do global actions
        try:
            if success:
                await self.task_success(job, stats=stats, stdout=stdout, stderr=stderr)
            else:
                future = None
                if reason:
                    stats['error_summary'] = reason
                    # check condor error for reset reason
                    for text in RESET_CONDOR_REASONS:
                        if text.lower() in reason.lower():
                            future = self.task_reset(job, stats=stats, reason=reason)
                            break
                if future is None and stderr and stderr.is_file():
                    # check stderr for reset reason
                    data = stderr.open().read()
                    for text in RESET_STDERR_REASONS:
                        if text.lower() in data.lower():
                            future = self.task_reset(job, stats=stats, reason=reason)
                            break
                if future is None:
                    future = self.task_failure(job, stats=stats, reason=reason, stdout=stdout, stderr=stderr)
                await future
        except Exception:
            logger.warning('failed to update REST', exc_info=True)

        # internal job cleanup
        del self.jobs[job_id]

    # Longer checks #

    async def check(self):
        """
        Do a cross-check, to verify `self.jobs` vs the submit dir and IceProd API.
        """
        logger.info('starting cross-check')

        all_jobs = self.submitter.get_jobs()

        # swap job dicts
        for job_id in set(self.jobs) - set(all_jobs):
            logger.info('removing job %s from cross-check', job_id)
        old_jobs = self.jobs
        self.jobs = all_jobs

        # process any updates
        for job_id, job in all_jobs.items():
            if job_id not in old_jobs or job.status != old_jobs[job_id].status:
                if job.status == JobStatus.FAILED:
                    extra = job.extra if job.extra else {}
                    reason = extra.get('HoldReason', 'Job has failed')
                    logger.info("job %s %s.%s removed from cross-check: %r", job_id, job.dataset_id, job.task_id, reason)
                    self.submitter.remove(job_id, reason=reason)
                else:
                    self.job_update(job)

        await self.check_history()

        await self.check_iceprod()

        # check for old jobs and dirs
        async for path in self.check_submit_dir():
            for job_id, job in self.jobs.items():
                if job.submit_dir == path:
                    self.submitter.remove(job_id, reason='exceeded max iceprod queue time')

        logger.info('finished cross-check')

    async def check_history(self):
        """Check condor_history"""
        now = time.time()
        hist_jobs = self.submitter.get_history(since=self.last_event_timestamp)
        self.last_event_timestamp = now
        for job_id, job in hist_jobs.items():
            if job_id not in self.jobs:
                self.jobs[job_id] = job

            logger.info("job %s %s.%s exited on its own from cross-check", job_id, job.dataset_id, job.task_id)
            extra = job.extra if job.extra else {}

            # get stats
            cpu = extra.get('CpusUsage', None)
            if not cpu and (wall := extra.get('LastRemoteWallClockTime', None)):
                cpu = (extra.get('RemoteSysCpu', 0) + extra.get('RemoteUserCpu', 0)) * 1. / wall
            gpu = extra.get('GpusUsage', None)
            memory = extra.get('ResidentSetSize_RAW', None)  # KB
            disk = extra.get('DiskUsage_RAW', None)  # KB
            time_ = extra.get('LastRemoteWallClockTime', None)  # seconds

            resources = {}
            if cpu is not None:
                resources['cpu'] = cpu
            if gpu is not None:
                resources['gpu'] = gpu
            if memory is not None:
                resources['memory'] = memory/1000000.
            if disk is not None:
                resources['disk'] = disk/1000000.
            if time_ is not None:
                resources['time'] = time_/3600.

            success = extra.get('JobStatus') == 4 and extra.get('ExitCode', 1) == 0
            job.status = JobStatus.COMPLETED if success else JobStatus.FAILED

            stats = {}
            if site := extra.get('MachineAttrGLIDEIN_Site0'):
                stats['site'] = site
            elif site := extra.get('MATCH_EXP_JOBGLIDEIN_ResourceName'):
                stats['site'] = site

            reason = None
            if r := extra.get('LastHoldReason'):
                reason = r
            elif r := extra.get('RemoveReason'):
                reason = r

            # finish job
            await self.finish(job_id, success=success, resources=resources, stats=stats, reason=reason)

    async def check_iceprod(self):
        """
        Sync with iceprod server status.
        """
        fut = self.get_tasks_on_queue()
        queue_tasks = {j.task_id: j for j in self.jobs.values()}
        server_tasks = await fut
        now = datetime.now(UTC)
        for task in server_tasks:
            if task['task_id'] not in queue_tasks:
                # ignore anything too recent
                if str2datetime(task['status_changed']) >= now - timedelta(minutes=1):
                    continue
                logger.info(f'task {task["dataset_id"]}.{task["task_id"]} in iceprod but not in queue')
                job = CondorJob(
                    dataset_id=task['dataset_id'],
                    task_id=task['task_id'],
                    instance_id=task['instance_id'],
                )
                await self.task_reset(job, reason='task missing from HTCondor queue')

    async def check_submit_dir(self):
        """
        Return directory paths that should be cleaned up.
        """
        # get time limits
        queue_tasks = {j.task_id for j in self.jobs.values()}
        queued_time = self.cfg['queue'].get('max_task_queued_time', 86400*2)
        processing_time = self.cfg['queue'].get('max_task_processing_time', 86400*2)
        suspend_time = self.cfg['queue'].get('suspend_submit_dir_time', 86400)
        now = time.time()
        job_clean_logs_time = now - suspend_time
        job_old_time = now - (queued_time + processing_time)
        dir_old_time = now - (queued_time + processing_time + suspend_time)
        logger.debug('now: %r, job_clean_logs_time: %r, job_old_time: %r, dir_old_time: %r', now, job_clean_logs_time, job_old_time, dir_old_time)

        for daydir in self.submit_dir.glob('[0-9][0-9][0-9][0-9]*'):
            logger.debug('looking at daydir %s', daydir)
            if daydir.is_dir():
                empty = True
                for path in daydir.iterdir():
                    job_active = path.name.split('_')[0] in queue_tasks
                    logger.debug('looking at path %s, active: %r', path, job_active)
                    st = path.lstat()
                    logger.debug('stat: %r', st)
                    if stat.S_ISDIR(st.st_mode):
                        empty = False
                        if not job_active:
                            if st.st_mtime < job_clean_logs_time:
                                logger.info('cleaning up submit dir %s', path)
                                shutil.rmtree(path)
                        elif st.st_mtime < job_old_time:
                            yield path
                            if st.st_mtime < dir_old_time:
                                logger.info('cleaning up submit dir %s', path)
                                shutil.rmtree(path)
                if empty:
                    logger.info('cleaning up daydir %s', daydir)
                    for path in self.jels.copy():
                        if Path(path).parent == daydir:
                            logger.info('removing JEL')
                            self.jels[path].close()
                            del self.jels[path]
                    shutil.rmtree(daydir)
                    continue
                # let other processing happen
                await asyncio.sleep(0)



"""
A supercomputer-like plugin, specificially for the SC 2019 Demo.

Basically, task submission directly to condor.
"""
import asyncio
from datetime import datetime,timedelta
from functools import partial
import getpass
import gzip
import logging
import os
import subprocess

import requests.exceptions

import iceprod
from iceprod.core import constants
from iceprod.core.exe_json import ServerComms
from iceprod.server import grid
from iceprod.server.globus import SiteGlobusProxy
from iceprod.server.plugins.condor import condor_os_reqs

logger = logging.getLogger('plugin-sc_demo')


async def check_call(*args, **kwargs):
    logger.info('subprocess_check_call: %r', args)
    p = await asyncio.create_subprocess_exec(*args, **kwargs)
    if p.returncode:
        raise Exception(f'command failed, return code {p.returncode}')
    return p


async def check_call_clean_env(*args, **kwargs):
    logger.info('subprocess_check_call: %r', args)
    env = os.environ.copy()
    del env['LD_LIBRARY_PATH']
    kwargs['env'] = env
    p = await asyncio.create_subprocess_exec(*args, **kwargs)
    if p.returncode:
        raise Exception(f'command failed, return code {p.returncode}')
    return p


async def check_output_clean_env(*args, **kwargs):
    logger.info('subprocess_check_output: %r', args)
    kwargs['stdout'] = subprocess.PIPE
    kwargs['stderr'] = subprocess.PIPE
    env = os.environ.copy()
    del env['LD_LIBRARY_PATH']
    kwargs['env'] = env
    p = await asyncio.create_subprocess_exec(*args, **kwargs)
    out,err = await p.communicate()
    out = out+b'\n'+err
    if p.returncode:
        raise Exception(f'command failed, return code {p.returncode}')
    return out.decode('utf-8')


class MyServerComms(ServerComms):
    def __init__(self, rest_client):
        self.rest = rest_client


class TaskInfo(dict):
    def __init__(self, **kwargs):
        self['dataset_id'] = None
        self['job_id'] = None
        self['task_id'] = None
        self['config'] = None
        self['pilot'] = None
        self['submit_dir'] = None
        super(TaskInfo, self).__init__(**kwargs)


def run_async(func,*args,**kwargs):
    async def myfunc():
        return_obj = {'args':args,'kwargs':kwargs}
        try:
            return_obj['return'] = await func(*args,**kwargs)
        except Exception as e:
            logger.info('error running async', exc_info=True)
            return_obj['exception'] = e
        return return_obj
    return asyncio.ensure_future(myfunc())


class sc_demo(grid.BaseGrid):
    """Plugin Overrides for SC Demo"""

    def __init__(self, *args, **kwargs):
        super(sc_demo, self).__init__(*args, **kwargs)
        self.x509proxy = SiteGlobusProxy()

        # SC demo queue requirements
        self.site = 'SC-Demo'
        if 'site' in self.queue_cfg:
            self.site = self.queue_cfg['site']
        self.resources = {'site': self.site}
        if 'gpu' in self.site.lower():
            self.resources['gpu'] = 1
        self.queue_params = {
            'requirements.site': self.site,
        }
        logger.info('resources: %r', self.resources)
        logger.info('queue params: %r', self.queue_params)

        self.grid_remove_once = set()

    async def upload_logfiles(self, task_id, dataset_id, submit_dir='', reason=''):
        """upload logfiles"""
        data = {'name': 'stdlog', 'task_id': task_id, 'dataset_id': dataset_id}

        def read_filename(filename):
            if os.path.exists(filename):
                with open(filename) as f:
                    return f.read()
            elif os.path.exists(filename+'.gz'):
                try:
                    with gzip.open(filename+'.gz', 'rt', encoding='utf-8') as f:
                        return f.read()
                except EOFError:
                    pass
                except Exception:
                    logging.info('stdlog:', exc_info=True)
            return ''

        # upload stdlog
        data['data'] = read_filename(os.path.join(submit_dir, constants['stdlog']))
        if not data['data']:
            data['data'] = reason
        await self.rest_client.request('POST', '/logs', data)

        # upload stderr
        data['name'] = 'stderr'
        data['data'] = read_filename(os.path.join(submit_dir, constants['stderr']))
        await self.rest_client.request('POST', '/logs', data)

        # upload stdout
        data['name'] = 'stdout'
        data['data'] = read_filename(os.path.join(submit_dir, constants['stdout']))
        await self.rest_client.request('POST', '/logs', data)

    async def task_error(self, task_id, dataset_id, submit_dir, reason=''):
        """reset a task"""
        # search for resources in stdout
        resources = {}
        filename = os.path.join(submit_dir, 'condor.out')
        if os.path.exists(filename):
            with open(filename) as f:
                resource_lines = False
                for line in f:
                    line = line.strip()
                    if line == 'Resources:':
                        resource_lines = True
                        continue
                    if resource_lines:
                        if ':' in line:
                            name,value = line.split(':',1)
                            if value.isdigit():
                                resources[name] = int(value)
                            else:
                                try:
                                    resources[name] = float(value)
                                except Exception:
                                    resources[name] = value
                            continue
                        else:
                            break
        if not reason:
            # search for reason in logfile
            filename = os.path.join(submit_dir, constants['stdlog'])
            if os.path.exists(filename):
                with open(filename) as f:
                    for line in f:
                        line = line.strip()
                        if 'failed to download' in line:
                            reason = 'failed to download input file(s)'
                        if 'failed to upload' in line:
                            reason = 'failed to upload output file(s)'
                        if 'Exception' in line and not reason:
                            reason = line
                        if 'return code:' in line and not reason:
                            reason = 'task error: return code '+line.rsplit(':',1)[-1]
        if not reason:
            reason = 'unknown failure'

        comms = MyServerComms(self.rest_client)
        await comms.task_error(task_id, dataset_id=dataset_id,
                               reason=reason, resources=resources,
                               site=self.site)

    async def finish_task(self, task_id, dataset_id, submit_dir):
        """complete a task"""
        # search for reasources in slurm stdout
        resources = {}
        filename = os.path.join(submit_dir, 'condor.out')
        if os.path.exists(filename):
            with open(filename) as f:
                resource_lines = False
                for line in f:
                    line = line.strip()
                    if line == 'Resources:':
                        resource_lines = True
                        continue
                    if resource_lines:
                        if ':' in line:
                            name,value = line.split(':',1)
                            if value.isdigit():
                                resources[name] = int(value)
                            else:
                                try:
                                    resources[name] = float(value)
                                except Exception:
                                    resources[name] = value
                            continue
                        else:
                            break

        comms = MyServerComms(self.rest_client)
        await comms.finish_task(task_id, dataset_id=dataset_id,
                                resources=resources)

    async def check_and_clean(self):
        """Check and clean the grid"""
        # return ### for queueing, don't need this
        host = grid.get_host()
        # self.x509proxy.update_proxy()

        # get time limits
        try:
            queued_time = timedelta(seconds=self.queue_cfg['max_task_queued_time'])
        except Exception:
            queued_time = timedelta(seconds=86400*2)
        try:
            processing_time = timedelta(seconds=self.queue_cfg['max_task_processing_time'])
        except Exception:
            processing_time = timedelta(seconds=86400*2)
        try:
            suspend_time = timedelta(seconds=self.queue_cfg['suspend_submit_dir_time'])
        except Exception:
            suspend_time = timedelta(seconds=86400)
        all_time = queued_time + processing_time + suspend_time
        time_dict = {'queued': queued_time,
                     'processing': queued_time+processing_time,
                     'completed': all_time,
                     'error': all_time,
                     'unknown': all_time}
        for t in time_dict:
            logger.debug("time limit: %s - %s",t,time_dict[t])
        now = datetime.utcnow()

        # get pilots from iceprod
        args = {
            'queue_host': host,
            'host': self.site,
            'keys': 'pilot_id|queue_host|grid_queue_id|submit_date|tasks',
        }
        ret = await self.rest_client.request('GET', '/pilots', args)

        # filter by queue host
        # index by grid_queue_id
        pilots = {}
        for pilot_id in ret:
            if (ret[pilot_id]['queue_host'] == host
                    and 'grid_queue_id' in ret[pilot_id]
                    and ret[pilot_id]['grid_queue_id']):
                pilots[ret[pilot_id]['grid_queue_id']] = ret[pilot_id]

        # first, check if anything has completed successfully
        # important to get grid status before completions, so we don't reset
        # newly completed jobs
        grid_jobs = await self.get_grid_status()
        grid_history = await self.get_grid_completions()

        logger.debug("iceprod pilots: %r", list(pilots))
        logger.debug("grid jobs: %r", list(grid_jobs))
        logger.debug("grid history: %r", list(grid_history))

        if grid_history:
            pilot_futures = []
            pilots_to_delete = {}
            for gid in pilots:
                if gid in grid_history:
                    pilot_id = None
                    try:
                        pilot = pilots[gid]
                        pilot_id = pilot['pilot_id']
                        pilot['submit_dir'] = grid_history[gid]['submit_dir']
                        if pilot['tasks']:
                            task_id = pilot['tasks'][0]
                            logger.info('post-processing task %s', task_id)
                            ret = await self.rest_client.request('GET', f'/tasks/{task_id}')
                            if ret['status'] == 'processing':
                                pilot['dataset_id'] = ret['dataset_id']
                                if grid_history[gid]['status'] == 'ok':
                                    logger.info('uploading logs for task %s', task_id)
                                    pilot_futures.append(asyncio.ensure_future(self.upload_output(pilot)))
                                else:
                                    logger.info('error in task %s', task_id)
                                    await self.upload_logfiles(task_id, pilot['dataset_id'],
                                                               submit_dir=pilot['submit_dir'])
                                    await self.task_error(task_id, pilot['dataset_id'],
                                                          submit_dir=pilot['submit_dir'])
                    except Exception:
                        logger.error('error handling task', exc_info=True)

                    if pilot_id:
                        logger.info('deleting completed pilot %s, with gid %s', pilot_id, gid)
                        pilots_to_delete[pilot_id] = gid

            for fut in asyncio.as_completed(pilot_futures):
                pilot,e = await fut  # upload is done

                try:
                    task_id = pilot['tasks'][0]
                    if e:
                        logger.info('failed to upload output files for task %s', task_id)
                        reason = f'failed to upload output files\n{e}'
                        await self.upload_logfiles(task_id,
                                                   dataset_id=pilot['dataset_id'],
                                                   submit_dir=pilot['submit_dir'],
                                                   reason=reason)
                        await self.task_error(task_id,
                                              dataset_id=pilot['dataset_id'],
                                              submit_dir=pilot['submit_dir'],
                                              reason=reason)
                    else:
                        logger.info('finishing task %s', task_id)
                        await self.upload_logfiles(task_id,
                                                   dataset_id=pilot['dataset_id'],
                                                   submit_dir=pilot['submit_dir'])
                        await self.finish_task(task_id,
                                               dataset_id=pilot['dataset_id'],
                                               submit_dir=pilot['submit_dir'])
                except Exception:
                    logger.error('error handling task', exc_info=True)

            for pilot_id in pilots_to_delete:
                try:
                    await self.rest_client.request('DELETE', f'/pilots/{pilot_id}')
                except Exception:
                    logger.info('delete pilot error', exc_info=True)
                gid = pilots_to_delete[pilot_id]
                if gid in pilots:
                    del pilots[gid]

        # Now do the regular check and clean
        reset_pilots = set(pilots).difference(grid_jobs)
        prechecked_dirs = set()

        # give two attempts to find a grid job before removing it
        remove_grid_jobs = set()
        remove_once = set()
        for gid in set(grid_jobs).difference(pilots):
            if gid in self.grid_remove_once:
                remove_grid_jobs.add(gid)
            else:
                remove_once.add(gid)
        self.grid_remove_once = remove_once

        # check the queue
        grid_idle = 0
        for grid_queue_id in set(grid_jobs).intersection(pilots):
            status = grid_jobs[grid_queue_id]['status']
            submit_time = pilots[grid_queue_id]['submit_date']
            if '.' in submit_time:
                submit_time = datetime.strptime(submit_time, '%Y-%m-%dT%H:%M:%S.%f')
            else:
                submit_time = datetime.strptime(submit_time, '%Y-%m-%dT%H:%M:%S')

            if now - submit_time > time_dict[status]:
                logger.info('pilot over time: %r', pilots[grid_queue_id]['pilot_id'])
                reset_pilots.add(grid_queue_id)
            elif status == 'queued':
                grid_idle += 1

            submit_dir = grid_jobs[grid_queue_id]['submit_dir']
            if submit_dir:
                # queueing systems don't like deleteing directories they know
                # about, so put them on a list of "don't touch"
                prechecked_dirs.add(submit_dir)
        self.grid_idle = grid_idle
        self.grid_processing = len(pilots)-len(reset_pilots)-grid_idle

        # check submit directories
        delete_dirs = set()
        for x in os.listdir(self.submit_dir):
            d = os.path.join(self.submit_dir,x)
            if d in prechecked_dirs:
                continue
            if os.path.isdir(d) and '_' in x:
                logger.debug('found submit_dir %s',d)
                mtime = datetime.utcfromtimestamp(os.path.getmtime(d))
                # use all_time instead of suspend_time because the
                # dir will have the submit time, not the last time
                if now-mtime < all_time:
                    continue  # skip for suspended or failed tasks
                delete_dirs.add(d)

        logger.info('%d processing pilots', self.grid_processing)
        logger.info('%d queued pilots', self.grid_idle)
        logger.info('%d ->reset', len(reset_pilots))
        logger.info('%d ->grid remove', len(remove_grid_jobs))
        logger.info('%d ->submit clean', len(delete_dirs))
        self.statsd.gauge('processing_pilots', self.grid_processing)
        self.statsd.gauge('queued_pilots', self.grid_idle)
        self.statsd.incr('reset_pilots', len(reset_pilots))
        self.statsd.incr('grid_remove', len(remove_grid_jobs))
        self.statsd.incr('clean_dirs', len(delete_dirs))

        if reset_pilots:
            logger.info('reset %r',reset_pilots)
            for grid_queue_id in reset_pilots:
                try:
                    pilot_id = pilots[grid_queue_id]['pilot_id']
                    await self.rest_client.request('DELETE', '/pilots/{}'.format(pilot_id))
                except KeyError:
                    pass
                except requests.exceptions.HTTPError as e:
                    if e.response.status_code == 404:
                        continue  # a missing pilot is the point of deleting it
                    logger.info('delete pilot error', exc_info=True)
                except Exception:
                    logger.info('delete pilot error', exc_info=True)

        # remove grid tasks
        if remove_grid_jobs:
            logger.info('remove %r',remove_grid_jobs)
            await asyncio.ensure_future(self.remove(remove_grid_jobs))

        if delete_dirs:
            await asyncio.ensure_future(self._delete_dirs(delete_dirs))

        # os._exit(0) # only run this once, then exit once cleanup is done

    async def queue(self):
        """Submit a pilot for each task, up to the limit"""
        host = grid.get_host()
        # self.x509proxy.update_proxy()

        dataset_cache = {}

        async def make_pilot(task):
            """take a TaskInfo as input"""
            resources = self.resources.copy()

            # create pilot
            if 'time' in resources:
                resources_available = {'time': resources['time']}
            else:
                resources_available = {'time': 1}
            for k in ('cpu','gpu','memory','disk'):
                if k in resources:
                    resources_available[k] = resources[k]-task['requirements'][k]
                else:
                    resources_available[k] = 0
            pilot = {
                'resources': resources,
                'resources_available': resources_available,
                'resources_claimed': task['requirements'],
                'tasks': [task['task_id']],
                'queue_host': host,
                'queue_version': iceprod.__version__,
                'host': self.site,
                'version': iceprod.__version__,
            }
            ret = await self.rest_client.request('POST', '/pilots', pilot)
            pilot['pilot_id'] = ret['result']
            task['pilot'] = pilot

            # get full job, dataset, config info
            job = await self.rest_client.request('GET', f'/jobs/{task["job_id"]}')
            if task['dataset_id'] in dataset_cache:
                dataset, config = dataset_cache[task['dataset_id']]
            else:
                dataset = await self.rest_client.request('GET', f'/datasets/{task["dataset_id"]}')
                config = await self.rest_client.request('GET', f'/config/{task["dataset_id"]}')
                dataset_cache[task['dataset_id']] = (dataset, config)

            # check if we have any files in the task_files API
            task_cfg = None
            for t in config['tasks']:
                if t['name'] == task['name']:
                    task_cfg = t
                    break
            else:
                raise Exception(f'cannot find task in config for {task["task_id"]}')

            if 'task_files' in task_cfg and task_cfg['task_files']:
                comms = MyServerComms(self.rest_client)
                files = await comms.task_files(task['dataset_id'],
                                               task['task_id'])
                task_cfg['data'].extend(files)

            config['dataset'] = dataset['dataset']
            task.update({
                'config': config,
                'job': job['job_index'],
                'dataset': dataset['dataset'],
                'jobs_submitted': dataset['jobs_submitted'],
                'tasks_submitted': dataset['tasks_submitted'],
                'debug': dataset['debug'],
                'requirements': self.resources.copy(),
            })
            await self.setup_submit_directory(task)
            task['pilot']['submit_dir'] = task['submit_dir']

            if any(f['movement'] in ('download','both') for f in task_cfg['data']):
                # get input files, all tasks in parallel
                await self.download_input(task['pilot'])

            # submit to queue
            await self.submit(task['pilot'])

            # update pilot
            pilot_id = task['pilot']['pilot_id']
            args = {'grid_queue_id': task['pilot']['grid_queue_id']}
            await self.rest_client.request('PATCH', f'/pilots/{pilot_id}', args)

        awaitables = set()
        for _ in range(self.get_queue_num()):
            # get a processing task
            args = {
                'requirements': self.resources.copy(),
                'query_params': self.queue_params,
            }
            args['requirements']['os'] = 'RHEL_7_x86_64'
            try:
                ret = await self.rest_client.request('POST', '/task_actions/process', args)
            except Exception:
                logger.info('no more tasks to queue')
                break
            awaitables.add(run_async(make_pilot, TaskInfo(**ret)))

        grid_queue_ids = []
        for fut in asyncio.as_completed(awaitables):
            ret = await fut
            task = ret['args'][0]

            if 'exception' in ret:
                reason = f'failed queue task\n{ret["exception"]}'
                logger.warning(reason)
                await self.upload_logfiles(task['task_id'],
                                           dataset_id=task['dataset_id'],
                                           submit_dir=task['submit_dir'],
                                           reason=reason)
                await self.task_error(task['task_id'],
                                      dataset_id=task['dataset_id'],
                                      submit_dir=task['submit_dir'],
                                      reason=reason)
                if task['pilot']['pilot_id']:
                    logger.info("deleting just submitted pilot: %s", task["pilot"]["pilot_id"])
                    await self.rest_client.request('DELETE', f'/pilots/{task["pilot"]["pilot_id"]}')
                if 'grid_queue_id' in task['pilot'] and task['pilot']['grid_queue_id']:
                    logger.info("deleting just submitted job: %s", task['pilot']['grid_queue_id'])
                    await self.remove([task['pilot']['grid_queue_id']])
            else:
                if 'grid_queue_id' in task['pilot'] and task['pilot']['grid_queue_id']:
                    grid_queue_ids.append(task['pilot']['grid_queue_id'])

        # DEMO: put jobs on hold, to releaes when it's time
        # cmd = ['condor_hold']+grid_queue_ids
        # out = await check_output_clean_env(*cmd)

    async def download_input(self, task):
        """
        Download input files for task.

        Args:
            task (dict): task info
        Returns:
            dict: task info
        """
        try:
            await check_call(
                'python', '-m', 'iceprod.core.data_transfer', '-f',
                os.path.join(task['submit_dir'],'task.cfg'),
                '-d', task['submit_dir'],
                'input'
            )
        except Exception:
            logger.info('error downloading', exc_info=True)
            raise

    async def upload_output(self, task):
        """
        Upload output files for task.

        Args:
            task (dict): task info
        Returns:
            dict: task info
        """
        try:
            await check_call(
                'python', '-m', 'iceprod.core.data_transfer', '-f',
                os.path.join(task['submit_dir'],'task.cfg'),
                '-d', task['submit_dir'],
                'output'
            )
        except Exception as e:
            logger.info('error uploading', exc_info=True)
            return (task,e)
        return (task,None)

    async def generate_submit_file(self, task, cfg=None, passkey=None,
                                   filelist=None):
        """Generate queueing system submit file for task in dir."""
        args = self.get_submit_args(task,cfg=cfg)
        args.extend(['--offline', '--gzip-logs'])

        # get requirements and batchopts
        requirements = []
        batch_opts = {}
        for b in self.queue_cfg['batchopts']:
            if b.lower() == 'requirements':
                requirements.append(self.queue_cfg['batchopts'][b])
            else:
                batch_opts[b] = self.queue_cfg['batchopts'][b]
        if cfg:
            if (cfg['steering'] and 'batchsys' in cfg['steering'] and
                    cfg['steering']['batchsys']):
                for b in cfg['steering']['batchsys']:
                    if b.lower().startswith(self.__class__.__name__):
                        # these settings apply to this batchsys
                        for bb in cfg['steering']['batchsys'][b]:
                            value = cfg['steering']['batchsys'][b][bb]
                            if bb.lower() == 'requirements':
                                requirements.append(value)
                            else:
                                batch_opts[bb] = value
            if 'task' in cfg['options']:
                t = cfg['options']['task']
                if t in cfg['tasks']:
                    alltasks = [cfg['tasks'][t]]
                else:
                    alltasks = []
                    try:
                        for tt in cfg['tasks']:
                            if t == tt['name']:
                                alltasks.append(tt)
                    except Exception:
                        logger.warning('error finding specified task to run for %r',
                                       task,exc_info=True)
            else:
                alltasks = cfg['tasks']
            for t in alltasks:
                if 'batchsys' in t and t['batchsys']:
                    for b in t['batchsys']:
                        if b.lower().startswith(self.__class__.__name__):
                            # these settings apply to this batchsys
                            for bb in t['batchsys'][b]:
                                value = t['batchsys'][b][bb]
                                if bb.lower() == 'requirements':
                                    requirements.append(value)
                                else:
                                    batch_opts[bb] = value

        # write the submit file
        submit_file = os.path.join(task['submit_dir'],'condor.submit')
        with open(submit_file,'w') as f:
            p = partial(print,sep='',file=f)
            p('universe = vanilla')
            p('executable = {}'.format(os.path.join(task['submit_dir'],'loader.sh')))
            p('log = condor.log')
            p('output = condor.out')
            p('error = condor.err')
            p('notification = never')
            p('+IsIceProdJob = True')  # mark as IceProd for monitoring
            p('want_graceful_removal = True')
            if filelist:
                p('transfer_input_files = {}'.format(','.join(filelist)))
            p('skip_filechecks = True')
            p('should_transfer_files = always')
            p('when_to_transfer_output = ON_EXIT_OR_EVICT')
            p('+SpoolOnEvict = False')
            p('transfer_output_files = iceprod_log.gz, iceprod_out.gz, iceprod_err.gz')

            # handle resources
            p('+JobIsRunning = (JobStatus =!= 1) && (JobStatus =!= 5)')
            if 'cpu' in task['requirements'] and task['requirements']['cpu']:
                p('request_cpus = {}'.format(task['requirements']['cpu']))
            if 'gpu' in task['requirements'] and task['requirements']['gpu']:
                p('request_gpus = {}'.format(task['requirements']['gpu']))
                # DEMO: gpu jobs need region set by condor_qedit
                # p('+CHUNK_Locations="NONE"')
                # requirements.append('stringListIMember(CLOUD_DATARegion, CHUNK_Locations)')
            else:
                p('+CPU_START=False')
                requirements.append('(!isUndefined(Target.GPUs) ? Target.GPUs == 0 : True) && CPU_START')
            if 'memory' in task['requirements'] and task['requirements']['memory']:
                p('request_memory = {}'.format(int(task['requirements']['memory']*1000+100)))
            if 'disk' in task['requirements'] and task['requirements']['disk']:
                p('request_disk = {}'.format(int(task['requirements']['disk']*1000000)))
            if 'time' in task['requirements'] and task['requirements']['time']:
                # extra 10 min for pilot
                p('+OriginalTime = {}'.format(int(task['requirements']['time'])*3600+600))
            if 'os' in task['requirements'] and task['requirements']['os']:
                requirements.append(condor_os_reqs(task['requirements']['os']))

            for b in batch_opts:
                p(f'{b}={batch_opts[b]}')
            if requirements:
                p('requirements = ('+')&&('.join(requirements)+')')

            p('arguments = ',' '.join(args))
            p('queue')

    async def submit(self,task):
        """Submit task to queueing system."""
        cmd = ['condor_submit','-terse','condor.submit']
        for tries in range(3):  # make three attempts
            try:
                out = await check_output_clean_env(*cmd, cwd=task['submit_dir'])
                break
            except Exception:
                if tries >= 2:
                    raise
                await asyncio.sleep(1)  # backoff a bit
        grid_queue_id = []
        for line in out.split('\n'):
            # look for range
            parts = [p.strip() for p in line.split('-') if p.strip()]
            if len(parts) != 2:
                continue
            major = parts[0].split('.')[0]
            minor_1 = int(parts[0].split('.')[1])
            minor_2 = int(parts[1].split('.')[1])
            for i in range(minor_1, minor_2+1):
                grid_queue_id.append('{}.{}'.format(major,i))
        task['grid_queue_id'] = ','.join(grid_queue_id)

        return task

    async def get_grid_status(self):
        """Get all tasks running on the queue system.
           Returns {grid_queue_id:{status,submit_dir}}
        """
        ret = {}
        cmd = ['condor_q',getpass.getuser(),'-af:j','jobstatus','cmd']
        out = await check_output_clean_env(*cmd)
        print('get_grid_status():',out)
        for line in out.split('\n'):
            if not line.strip():
                continue
            parts = line.split()
            gid = parts[0]
            status = parts[1]
            cmd = parts[-1]
            if 'loader.sh' not in cmd:
                continue
            if status in ('0', '1', '5'):  # DEMO: treat held jobs as queued
                status = 'queued'
            elif status in ('2', '3', '6', '7'):
                status = 'processing'
            elif status == '4':
                status = 'completed'
            elif status in ('5',):
                status = 'error'
            else:
                status = 'unknown'
            ret[gid] = {'status':status,'submit_dir':os.path.dirname(cmd)}
        return ret

    async def remove(self,tasks):
        """Remove tasks from queueing system."""
        if tasks:
            cmd = ['condor_rm']+list(tasks)
            await check_call_clean_env(*cmd)

    async def get_grid_completions(self):
        """
        Get completions in the last 4 days.

        Returns:
            dict: {grid_queue_id: {status, submit_dir} }
        """
        ret = {}
        cmd = ['condor_history',getpass.getuser(),'-match','50000','-af:j','jobstatus','exitcode','exitbysignal','cmd']
        out = await check_output_clean_env(*cmd)
        print('get_grid_completions():',out)
        for line in out.split('\n'):
            if not line.strip():
                continue
            gid,status,exitstatus,exitsignal,cmd = line.split()
            if 'loader.sh' not in cmd:
                continue
            if status == '4' and exitstatus == '0' and exitsignal == 'false':
                status = 'ok'
            else:
                status = 'error'
            ret[gid] = {'status':status,'submit_dir':os.path.dirname(cmd)}
        return ret



"""
Mark jobs as completed.

Check tasks in a job, and when all are complete mark the job as complete.
"""

import argparse
import asyncio
import logging

from iceprod.client_auth import add_auth_to_argparse, create_rest_client

logger = logging.getLogger('job_completion')


async def run(rest_client, dataset_id=None, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        dataset_id (str): specific dataset to run on
        debug (bool): debug flag to propagate exceptions
    """
    if dataset_id:
        dataset_ids = [dataset_id]
    else:
        datasets = await rest_client.request('GET', '/dataset_summaries/status')
        dataset_ids = datasets.get('processing', [])
    for dataset_id in dataset_ids:
        try:
            jobs = await rest_client.request('GET', '/datasets/{}/job_summaries/status'.format(dataset_id))
            for job_id in jobs.get('processing', []):
                tasks = await rest_client.request('GET', '/datasets/{}/tasks?keys=task_id|task_index|status&job_id={}'.format(dataset_id, job_id))
                task_statuses = set(task['status'] for task in tasks.values())
                if task_statuses == {'complete'}:
                    logger.info('dataset %s job %s status -> complete', dataset_id, job_id)
                    args = {'status': 'complete'}
                    await rest_client.request('PUT', '/datasets/{}/jobs/{}/status'.format(dataset_id,job_id), args)
                elif not task_statuses - {'complete', 'suspended'}:
                    logger.info('dataset %s job %s status -> suspended', dataset_id, job_id)
                    args = {'status': 'suspended'}
                    await rest_client.request('PUT', '/datasets/{}/jobs/{}/status'.format(dataset_id,job_id), args)
                elif not task_statuses - {'complete', 'failed', 'suspended'}:
                    logger.info('dataset %s job %s status -> errors', dataset_id, job_id)
                    args = {'status': 'errors'}
                    await rest_client.request('PUT', '/datasets/{}/jobs/{}/status'.format(dataset_id,job_id), args)
        except Exception:
            logger.error('error completing a job in dataset %s', dataset_id, exc_info=True)
            if debug:
                raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')
    parser.add_argument('--dataset-id', help='specific dataset to run on')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    asyncio.run(run(rest_client, dataset_id=args.dataset_id, debug=args.debug))


if __name__ == '__main__':
    main()
